;;; -*- Mode:Scheme; Base:10 -*- PS9-MODS.SCM

;;		     MASSACHUSETTS INSTITUTE OF TECHNOLOGY
;;	   Department of Electrical Engineering and Computer Science
;;	   6.001---Structure and Interpretation of Computer Programs
;;			   Spring Semester, 1993
;;
;;				 Problem Set 9


;;;  These procedures will be used when traced procedures are added to 
;;;  the evaluator.  The first is a modified version of USER-PRINT that 
;;;  calls the procedures you are to implement.  The original version 
;;;  is in the PS9-ECEV.SCM code.


(define (user-print object)
  (cond ((compound-procedure? object)
         (write-line (list 'compound-procedure
                      (parameters object)
		      (procedure-body object)
                      '[procedure-env])))
        ((compiled-procedure? object)
         (write-line '[compiled-procedure]))
; the following three lines will be used when tracing is added
        ((traced-procedure? object)			
         (write-line 'traced-procedure)
         (user-print (procedure-part-traced object)))
        (else (write-line object))))

;;;  This procedure will be helpful in actually printing the tracing 
;;;  information.

(define (print-call fun-name arglist)
	(write-line (list 'calling 
		     'procedure
		     fun-name
		     'with 
		     (reverse arglist))))       ; arguments stored in reverse


;;;; The following code is a copy of the tail-recursive EVAL-SEQUENCE

eval-sequence
  (assign exp (first-exp (fetch unev)))
  (branch (last-exp? (fetch unev)) last-exp)
  (save unev)
  (save env)
  (assign continue eval-sequence-continue)
  (goto eval-dispatch)
eval-sequence-continue
  (restore env)
  (restore unev)
  (assign unev (rest-exps (fetch unev)))
  (goto eval-sequence)
last-exp
  (restore continue)
  (goto eval-dispatch)

;;;; The following code is a non-tail-recursive version of EVAL-SEQUENCE

eval-sequence
  (branch (no-more-exps? (fetch unev)) end-sequence)
  (assign exp (first-exp (fetch unev)))
  (save unev)
  (save env)
  (assign continue eval-sequence-continue)
  (goto eval-dispatch)
eval-sequence-continue
  (restore env)
  (restore unev)
  (assign unev (rest-exps (fetch unev)))
  (goto eval-sequence)
end-sequence
  (restore continue)
  (goto (fetch continue))

;;;; The complete definition of Explicit-Control Evaluator register machine 

(define-machine explicit-control-evaluator
  (registers exp env val continue fun argl unev)
  (controller
  (perform (newline))
  (perform (display "========== Entering EC Eval REPL =========="))
read-eval-print-loop
  (perform (initialize-stack))			       ;;;
  (perform (initialize-ops-counter))                   ;;; 
  (perform (newline))
  (perform (display "----------------------------------------"))
  (perform (newline))
  (perform (display "EC-EVAL==> "))
  (assign exp (read))
  (perform (initialize-graphics-window))	       ;;;
  (branch (exit-on? (fetch exp)) ec-eval-exit)
  (assign env the-global-environment)
  (assign continue print-result)
  (goto eval-dispatch)
print-result
  (perform (user-print (fetch val)))
  (goto read-eval-print-loop)

unknown-procedure-type-error
  (assign val '>>>Unknown-procedure-type-error)
  (goto signal-error)
unknown-expression-type-error
  (assign val '>>>Unknown-expression-type-error)
signal-error
  (perform (newline))
  (perform (user-print (fetch val)))		;; Err mesg in VAL
  (perform (user-print (fetch exp)))		;; Irritant in EXP
  (goto read-eval-print-loop)

external-entry
   (perform (initialize-stack))
   (assign env the-global-environment)
   (assign continue print-result)
   (save continue)
   (goto (fetch val))

eval-dispatch
  (branch (self-evaluating?  (fetch exp)) ev-self-eval)
  (branch (quoted?           (fetch exp)) ev-quote)
  (branch (variable?         (fetch exp)) ev-variable)
  (branch (definition?       (fetch exp)) ev-definition)
  (branch (assignment?       (fetch exp)) ev-assignment)
  (branch (lambda?           (fetch exp)) ev-lambda)
  (branch (conditional?      (fetch exp)) ev-cond)
  (branch (no-args?          (fetch exp)) ev-no-args)
  (branch (args-application? (fetch exp)) ev-application)
  (goto unknown-expression-type-error)
ev-self-eval
  (assign val (fetch exp))
  (goto (fetch continue))
ev-quote
  (assign val (text-of-quotation (fetch exp)))
  (goto (fetch continue))
ev-variable
  (assign val (lookup-variable-value (fetch exp) (fetch env)))
  (goto (fetch continue))
ev-lambda
  (assign val (make-procedure (fetch exp) (fetch env)))
  (goto (fetch continue))
ev-no-args
  (assign exp (operator (fetch exp)))
  (save continue)
  (assign continue setup-no-arg-apply)
  (goto eval-dispatch)
setup-no-arg-apply
  (assign fun (fetch val))
  (assign argl '())
  (goto apply-dispatch)

ev-application
  (assign unev (operands (fetch exp)))
  (assign exp (operator (fetch exp)))
  (save continue)
  (save env)
  (save unev)
  (assign continue eval-args)
  (goto eval-dispatch)
eval-args
  (restore unev)
  (restore env)
  (assign fun (fetch val))
  (save fun)
  (assign argl '())
  (goto eval-arg-loop)

eval-arg-loop
  (save argl)
  (assign exp (first-operand (fetch unev)))
  (branch (last-operand? (fetch unev)) eval-last-arg)
  (save env)
  (save unev)
  (assign continue accumulate-arg)
  (goto eval-dispatch)
accumulate-arg
  (restore unev)
  (restore env)
  (restore argl)
  (assign argl (cons (fetch val) (fetch argl)))
  (assign unev (rest-operands (fetch unev)))
  (goto eval-arg-loop)
eval-last-arg
  (assign continue accumulate-last-arg)
  (goto eval-dispatch)
accumulate-last-arg
  (restore argl)
  (assign argl (cons (fetch val) (fetch argl)))
  (restore fun)
  (goto apply-dispatch)

apply-dispatch
  (branch (primitive-procedure? (fetch fun)) primitive-apply)
  (branch ( compound-procedure? (fetch fun))  compound-apply)
  (branch ( compiled-procedure? (fetch fun))  compiled-apply)
  (assign exp (fetch fun))		;; Place irritant in EXP
  (goto unknown-procedure-type-error)
compiled-apply
   (assign val (compiled-procedure-entry (fetch fun)))
   (goto (fetch val))
primitive-apply
  (assign val
          (apply-primitive-procedure (fetch fun)
                                     (fetch argl)))
  (restore continue)
  (goto (fetch continue))
compound-apply
  (assign env (make-bindings (fetch fun) (fetch argl)))
  (assign unev (procedure-body (fetch fun)))
  (goto eval-sequence)
eval-sequence
  (assign exp (first-exp (fetch unev)))
  (branch (last-exp? (fetch unev)) last-exp)
  (save unev)
  (save env)
  (assign continue eval-sequence-continue)
  (goto eval-dispatch)
eval-sequence-continue
  (restore env)
  (restore unev)
  (assign unev (rest-exps (fetch unev)))
  (goto eval-sequence)
last-exp
  (restore continue)
  (goto eval-dispatch)

ev-cond
  (save continue)
  (assign continue evcond-decide)
  (assign unev (clauses (fetch exp)))
evcond-pred
  (branch (no-clauses? (fetch unev)) evcond-return-nil)
  (assign exp (first-clause (fetch unev)))
  (branch (else-clause? (fetch exp)) evcond-else-clause)
  (save env)
  (save unev)
  (assign exp (predicate (fetch exp)))
  (goto eval-dispatch)

evcond-return-nil
  (restore continue)
  (assign val nil)
  (goto (fetch continue))
evcond-decide
  (restore unev)
  (restore env)
  (branch (true? (fetch val)) evcond-true-predicate)
  (assign unev (rest-clauses (fetch unev)))
  (goto evcond-pred)
evcond-true-predicate
  (assign exp (first-clause (fetch unev)))
evcond-else-clause
  (assign unev (actions (fetch exp)))
  (goto eval-sequence)
ev-assignment
  (assign unev (assignment-variable (fetch exp)))
  (save unev)
  (assign exp (assignment-value (fetch exp)))
  (save env)
  (save continue)
  (assign continue ev-assignment-1)
  (goto eval-dispatch)
ev-assignment-1
  (restore continue)
  (restore env)
  (restore unev)
  (perform
   (set-variable-value! (fetch unev) (fetch val) (fetch env)))
  (goto (fetch continue))
ev-definition
  (assign unev (definition-variable (fetch exp)))
  (save unev)
  (assign exp (definition-value (fetch exp)))
  (save env)
  (save continue)
  (assign continue ev-definition-1)
  (goto eval-dispatch)
ev-definition-1
  (restore continue)
  (restore env)
  (restore unev)
  (perform
   (define-variable! (fetch unev) (fetch val) (fetch env)))
  (assign val the-undefined-value)
  (goto (fetch continue))

ec-eval-exit
  (perform (newline))
  (perform (display "********** Exiting EC Eval REPL **********"))
  (assign val (exit (fetch exp)))	;; Leave exit message in VAL
  ))



